/**
 * gaim-write - Unix Write over SSH Protocol Plugin
 * http://sccs.swarthmore.edu/~cbr/gaim-write/
 * Copyright (C) 2005, Jeff Kaufman <cbr@sccs.swarthmore.edu>
 *
 * Built off:
 *   gaim - Null Protocol Plugin
 *   http://snarfed.org/space/gaim+null+protocol+plugin
 *   Copyright (C) 2004, Ryan Barrett <nullprpl@ryanb.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define GAIM_PLUGINS

#define COMPLACENT 0
#define WARY 1
#define PARINOID 2
#define TRUSTLEVEL  COMPLACENT


#include "gaim.h"
#include "internal.h"
#include "prpl.h"
#include "debug.h"
#include "version.h"
/*  #include "/usr/local/include/libssh/libssh.h"  */
/*  #include "/usr/local/include/libssh/priv.h"    */
#include "/usr/local/src/libssh-0.11/include/libssh/libssh.h"
#include "/usr/local/src/libssh-0.11/include/libssh/priv.h"
#include "gaimwrite.h"

static GaimPlugin *my_protocol = NULL;

char* G_WRITE_CONV_KEY = "GAIM_WRITE_CONV_KEY";

static const char* gaimwrite_list_icon(GaimAccount *a, GaimBuddy *b)
{
	return "gaimwrite";
}

static void gaimwrite_login(GaimAccount *account)
{
   gaim_debug_misc("gaimwrite",
		   "Preparing to log %s in.\n",gaim_account_get_username(account));

   GaimConnection *gc = gaim_account_get_connection(account);
   gaim_connection_set_state(gc, GAIM_CONNECTING);
   gaim_connection_update_progress(gc, _("Connecting..."), 0, 11);

   const char* user_at_host = gaim_account_get_username(account);
   int i;

   int host_begin = 0;
   int host_end = 0;

   /* break user_at_host into username and hostname segments */
   for(i = 0 ; 1 ; i++)
   {
      if(user_at_host[i] == '\0')
      {
	 if (host_begin)
	 {
	    host_end = i;
	    break;
	 }
	 else
	 {
	    gaim_connection_error(gc, _("Invalid username.  Want form \"user@host\"."));
	    return;
	 }
      }
      else if(user_at_host[i] == '@')
      {
	 if(host_begin)
	 {
	    gaim_connection_error(gc, _("Invalid username.  Want form \"user@host\"."));
	    return;
	 }
	 host_begin = i + 1;
      }
   }

   char username[host_begin];
   char hostname[host_end-host_begin+1];

   memcpy(username,user_at_host,host_begin-1);
   memcpy(hostname,user_at_host + host_begin,host_end-host_begin);

   username[host_begin] = '\0';
   hostname[host_end-host_begin] = '\0';


   gaim_debug_misc("gaimwrite", "Gaim-Write logging %s in to %s!\n",username,hostname);

   /* begin libssh code */

   SSH_OPTIONS* opts = options_new();
   options_set_port(opts, 22);
   options_set_host(opts, hostname);
   options_set_username(opts, username);

   if(options_set_wanted_method(opts,KEX_HOSTKEY,"ssh-dss,ssh-rsa"))
   {
      gaim_connection_error(gc,_("Encryption methods not properly supported"));
      return;
   }

   SSH_SESSION* ses = ssh_connect(opts);

   if (ses == NULL)
   {
      gaim_connection_error(gc,_("Connection to server failed"));
      return;
   }

   gaim_connection_update_progress(gc, _("Connected to the server over ssh."), 1, 11);
   gaim_debug_misc("gaimwrite","Connected to the server over ssh.\n");
   gaim_connection_update_progress(gc, _("Verifying server..."), 2, 11);
   
   switch(ssh_is_server_known(ses))
   {
      case SSH_SERVER_ERROR:
         gaim_connection_error(gc,_("Can't open known-hosts file."));
      case SSH_SERVER_NOT_KNOWN:
      case SSH_SERVER_FOUND_OTHER:
         gaim_debug_warning("gaimwrite","Server not known.\n");
         if(TRUSTLEVEL == PARINOID)
            gaim_connection_error(gc,_("Refusing to talk to an unknown host."));
         else if (TRUSTLEVEL == COMPLACENT)
         {
            gaim_debug_misc("gaimwrite","Adding host %s to known-hosts file.\n",hostname);
            ssh_write_knownhost(ses);
         }
	 else
	    gaim_debug_misc("gaimwrite","Accepting for this time only.\n");
         break;
      case SSH_SERVER_KNOWN_CHANGED:
         gaim_debug_warning("gaimwrite","WARNING!!! : The key for this host has changed. Someone may be trying to do something nasty.  Or maybe there was an upgrade or something.\n");
         if(TRUSTLEVEL == PARINOID || TRUSTLEVEL == WARY)
            gaim_connection_error(gc,_("Refusing to talk to a modified host."));
         else
	    gaim_debug_misc("gaimwrite","Accepting the modified key of a host.\n");
	 break;
      default:
         break;
   }   

   gaim_connection_update_progress(gc, _("Server accepted as legit."), 3, 11);
   gaim_debug_misc("gaimwrite","Server accepted as legit.\n");
   gaim_connection_update_progress(gc, _("Checking password..."), 4, 11);
   
   if(ssh_userauth_password(ses,username,gaim_account_get_password(account)))
   {
      gaim_connection_error(gc,_("Password not accepted."));
      return;
   }

   gaim_connection_update_progress(gc, _("Password acccepted"), 5, 11);
   gaim_debug_misc("gaimwrite","Password acccepted.\n");
   gaim_connection_update_progress(gc, _("Opening channel..."), 6, 11);
   
   /* now we're connected */

   CHANNEL* initial = channel_open_session(ses);

   if(initial == NULL)
   {
      gaim_connection_error(gc,_("Couldn't open a channel."));
      return;
   }

   gaim_connection_update_progress(gc, _("Channel opened."), 7, 11);
   gaim_debug_misc("gaimwrite","Channel opened.\n");
   gaim_connection_update_progress(gc, _("Opening pty..."), 8, 11);
   if(channel_request_pty(initial))
   {
      gaim_connection_error(gc,_("Couldn't open a pty."));
      return;
   }

   if(channel_request_shell(initial))
   {
      gaim_connection_error(gc,_("Couldn't get a shell."));
      return;
   }


   GWriteConn* conn = g_new0(GWriteConn, 1);

   conn->ssh_buffer = buffer_new();

   conn->ssh_session = ses;
   conn->listening = initial;

   gc->inpa = gaim_input_add( ses->fd , GAIM_INPUT_READ , gaimwrite_full_listener,gc);

   gc->proto_data = conn;

   conn->tosend = g_queue_new();

   gaimwrite_send_message(initial, "stty -echo\n",gc);
   gaimwrite_send_message(initial, "PS1=\n",gc);
   gaimwrite_send_message(initial, "mesg y\n",gc);

   gaim_connection_update_progress(gc, _("Pty opened."), 9, 11);
   gaim_debug_misc("gaimwrite","Pty opened.\n");
   gaim_connection_update_progress(gc, _("Connected"), 10, 11);
   gaim_connection_set_state(gc, GAIM_CONNECTED);


   conn->callback_handle = gaim_timeout_add(1000,send_next_queued_message,conn);

   serv_finish_login(gc);
}

static void gaimwrite_send_message(CHANNEL* chan, char* message,GaimConnection* gc)
{
   gaim_debug_misc("gaimwrite","Asked to send message \"%s\"\n",message);

   char* msgcpy = (char*)malloc(strlen(message));
   strcpy(msgcpy,message);

   gaimwrite_queue_message(G_WRITE_CONN(gc->proto_data), gaimwrite_new_sender(chan,msgcpy));
}

static GWriteToSend* gaimwrite_new_sender(CHANNEL* chan, char* message)
{
   GWriteToSend* ts = g_new0(GWriteToSend,1);
   ts->chan = chan;
   ts->message = message;
   ts->action = G_WRITE_SEND_MESSAGE;
   return ts;
}

static GWriteToSend* gaimwrite_new_closer(CHANNEL* chan) 
{ 
   GWriteToSend* ts = g_new0(GWriteToSend,1); 
   ts->chan = chan; 
   ts->message = NULL;
   ts->action = G_WRITE_REMOVE_CHANNEL; 
   return ts; 
} 


static void gaimwrite_queue_message(GWriteConn* conn, GWriteToSend* ts)
{
   g_queue_push_head(conn->tosend,ts);
}

static gboolean send_next_queued_message(gpointer data)
{
   GQueue* tosend = G_WRITE_CONN(data)->tosend;
   if(tosend != NULL && !g_queue_is_empty(tosend))
   {
      gpointer dat = g_queue_pop_tail(tosend);
      GWriteToSend* ts = G_WRITE_TO_SEND(dat);

      if(ts->action == G_WRITE_SEND_MESSAGE)
      {
	 int written = channel_write(ts->chan,ts->message,strlen(ts->message));
	 gaim_debug_misc("gaimwrite","Wrote %d out of %d bytes of \"%s\".\n",
			 written,strlen(ts->message),ts->message);
	 free(dat);
	 free(ts->message);
      }
      else if(ts->action == G_WRITE_REMOVE_CHANNEL)
      {
	 gaim_debug_misc("gaimwrite","Closing channel.\n");
	 gaimwrite_close_channel(ts->chan);
      }
      else
	 gaim_debug_misc("gaimwrite","Unknown action %d in send_next_queued_message\n",ts->action);
   }

   return TRUE;
}

static void gaimwrite_full_listener(gpointer data,gint source, GaimInputCondition cond)
{

   gaim_debug_misc("gaimwrite","Heard there was something to look for.  Commencing looking.\n");

   GaimConnection* gc = (GaimConnection*)data;
   GWriteConn* conn = G_WRITE_CONN(gc->proto_data);


   /* check the main line */

   CHANNEL* chan = conn->listening; 

/*   gaim_debug_misc("gaimwrite","A\n");*/
   char* message = gaimwrite_read_nb_channel(chan,gc);
/*   gaim_debug_misc("gaimwrite","B\n");*/


   if(message != NULL)
   { 
      gaim_debug_misc("gaimwrite","Got some data on the main channel: "); 

      /* need to parse recieved data, but for now we'l just print it */
      gaim_debug_misc("gaimwrite","%s\n",message); 
   } 
   
   
   gaim_debug_misc("gaimwrite","Moving on to check the conversation channels.\n");

   /* now we need to check all the channels to see if we have
    * anything new to report */

   GList* all_gconvs = gaim_get_conversations();
   GaimConversation* gconv = NULL;

   while(all_gconvs != NULL)
   {
      gconv = (GaimConversation*)all_gconvs->data;
      if(gconv->account == gaim_connection_get_account(gc)) /* only consider our conversations */
      {
	 GWriteConv* convo = gaimwrite_get_convo(gconv);
	 chan = convo->chan;

/*	 gaim_debug_misc("gaimwrite","C\n");*/
	 message = gaimwrite_read_nb_channel(chan,gc);
/*	 gaim_debug_misc("gaimwrite","D\n");*/

	 
	 if(message != NULL)
	 {
            serv_got_im(gc, gconv->name, message, 0, time(NULL));
	    gaim_debug_misc("gaimwrite","  Message: %s.\n",message);
	 }
      }
      all_gconvs = all_gconvs->next;
   }
   gaim_debug_misc("gaimwrite","Done processing fd change\n");
}

/* reads from the channel, strips unprintables and trailing newlines, makes
 * sure it is null-terminated, and returns NULL on failure */
static char* gaimwrite_read_nb_channel(CHANNEL* chan,GaimConnection* gc)
{
   char* locbuf = G_WRITE_CONN(gc->proto_data)->locbuf;
   int i = channel_read_nonblocking(chan,locbuf,8192,0);
   if(i > 0)
      return gaimwrite_clean_chars(locbuf, i);
   return NULL;
}

static char* gaimwrite_clean_chars(char* c, int end)
{
   if(end >= 8192)
      end = 8191;
   
   /* remove control characters */
   int j;
   for (j = 0 ; j < end ; j++)
   {
      if ( !((c[j] >= 32 && c[j] <= 126) 
	     || (c[j] == '\n') 
	     || (c[j] == 13)))
      {
	 gaim_debug_misc("gaimwrite","char %d unprintable : char %d\n",j,c[j]);
	 c[j] = '*';
      }
   }

   /* make sure it is null terminated */
   c[end] = 0;

   
   /* remove trailing newlines */
   j = end-1;
   while(j>=0 && (c[j] == '\n' || c[j] == 13 || c[j] == 10))
      c[j--] = 0;

   return c;
}

static void gaimwrite_close(GaimConnection* gc)
{
   gaim_debug_misc("gaimwrite","Preparing to disconnect from server.\n");

   GWriteConn* conn = G_WRITE_CONN(gc->proto_data);
   gaim_timeout_remove(conn->callback_handle);

   gaim_input_remove(gc->inpa);

   free(conn->locbuf);

   channel_send_eof(conn->listening);
   channel_free(conn->listening);
   ssh_disconnect(conn->ssh_session);
   gaim_debug_misc("gaimwrite","Disconnected ssh session.\n");

   gaim_debug_misc("gaimwrite","Server disconnected.\n");
}

static void gaimwrite_add_buddy(GaimConnection *gc, GaimBuddy *buddy,
				GaimGroup *group)
{
	gaim_debug_misc("gaimwrite", "Gaim-Write adding buddy: %s", buddy->name);
	serv_got_update(gc, buddy->name, 1, 0, 0, 0, 0);
}

static void gaimwrite_remove_buddy(GaimConnection *gc, GaimBuddy *buddy,
				   GaimGroup *group)
{

	gaim_debug_misc("gaimwrite", "Gaim-Write removing buddy: %s", buddy->name);
}

static int gaimwrite_setup_conv(GaimConnection* gc,
				GaimConversation* gconv,
				GWriteConn* conn,
				GWriteConv* convo,
				char* who,
				char* message  )
{
   conn = G_WRITE_CONN(gc->proto_data);
   CHANNEL* chan = channel_open_session(conn->ssh_session);
	 
   if(chan == NULL)
   {
      gaim_connection_error(gc,_("Couldn't open a channel."));
      return 0;
   }
	 
   gaim_debug_misc("gaimwrite","Channel opened to %s.\n",who);
   if(channel_request_pty(chan))
   {
      gaim_connection_error(gc,_("Couldn't open a pty."));
      return 0;
   }
	 
   if(channel_request_shell(chan))
   {
      gaim_connection_error(gc,_("Couldn't get a shell."));
      return 0;
   } 

   convo = g_new0(GWriteConv, 1);
   convo->chan = chan;
   gaim_conversation_set_data(gconv,G_WRITE_CONV_KEY,convo);

   gaimwrite_send_message(chan,"stty -echo\n",gc);
   gaimwrite_send_message(chan,"PS1=\n",gc);
   gaimwrite_send_message(chan,"write ",gc);
   gaimwrite_send_message(chan,who,gc);
   gaimwrite_send_message(chan,"\n",gc);

   gaim_debug_misc("gaimwrite", "sent initialiazation commands.\n");

   gaimwrite_send_message(chan,message,gc); 
   gaimwrite_send_message(chan,"\n",gc);

   return 1;
}
	 

static int gaimwrite_send_im(GaimConnection *gc, const char *who,
			     const char *message, GaimConvImFlags flags)
{

   /* first figure out if we have an active conversation with them */

   GaimConversation* gconv = gaim_find_conversation_with_account(who,gc->account);
   GWriteConn* conn;
   GWriteConv* convo;

   if (gconv != NULL)
   {
      void* tmpdata = gaim_conversation_get_data(gconv,G_WRITE_CONV_KEY);
      if (tmpdata != NULL)
      {
	 convo = G_WRITE_CONV(tmpdata);
	 gaim_debug_misc("gaimwrite","Found prior gconv with %s and data was set.\n",who);

/*	 if(convo->have_recieved)       */

	 gaimwrite_send_message(convo->chan,message,gc);
	 gaimwrite_send_message(convo->chan,"\n",gc);
	 return 1;
      }
      else
      {
	 gaim_debug_misc("gaimwrite","Found prior gconv with %s but data was unset.\n",who);
	 return gaimwrite_setup_conv(gc,gconv,conn,convo,who,message);
      }
   }

   gaim_debug_error("gaimwrite","No GaimConversation with %s found.\n",who); 
   return 0;
}

static GWriteConv* gaimwrite_get_convo(GaimConversation* gconv)
{
  void* tmpdata = gaim_conversation_get_data(gconv,G_WRITE_CONV_KEY); 
  if (tmpdata == NULL)
     return NULL;
  return G_WRITE_CONV(tmpdata);
}

static void gaimwrite_close_channel(CHANNEL* chan)
{
     gaim_debug_misc("gaimwrite","Freeing channel...\n");
     channel_send_eof(chan);
     channel_free(chan);
     gaim_debug_misc("gaimwrite","Channel freed.\n");
}

static void gaimwrite_close_convo(GaimConnection* gconn,const char *who)
{
   gaim_debug_misc("gaimwrite","Closing conversation with %s.\n",who);
   GaimConversation* gconv = gaim_find_conversation_with_account(who,gconn->account);

   GWriteConv* convo = gaimwrite_get_convo(gconv);

   if (convo != NULL) 
   {
      gaimwrite_queue_message(G_WRITE_CONN(gconn->proto_data), gaimwrite_new_closer(convo->chan)); 
   }
   else
   {
      gaim_debug_misc("gaimwrite","Nothing to do in close convo.\n");
   }
}

static GList *gaimwrite_away_states(GaimConnection *gc)
{
	GList *m = NULL;

	m = g_list_append(m, GAIM_AWAY_CUSTOM);
	m = g_list_append(m, _("Back"));

	return m;
}

static void gaimwrite_set_away(GaimConnection *gc, const char *state, const char *text)
{

}

static GaimPluginProtocolInfo prpl_info =
{
	0,
	NULL,					/* user_splits */
	NULL,					/* protocol_options */
	NO_BUDDY_ICONS,			/* icon_spec */
	gaimwrite_list_icon,		/* list_icon */
	NULL,					/* list_emblems */
	NULL,					/* status_text */
	NULL,					/* tooltip_text */
	gaimwrite_away_states,       /* away_states */
	NULL,					/* blist_node_menu */
	NULL,					/* chat_info */
	NULL,					/* chat_info_defaults */
	gaimwrite_login,			/* login */
	gaimwrite_close,					/* close */
	gaimwrite_send_im,		/* send_im */
	NULL,					/* set_info */
	NULL,					/* send_typing */
	NULL,					/* get_info */
	gaimwrite_set_away,		/* set_away */
	NULL,					/* set_idle */
	NULL,					/* change_passwd */
	gaimwrite_add_buddy,		/* add_buddy */
	NULL,					/* add_buddies */
	gaimwrite_remove_buddy,	/* remove_buddy */
	NULL,					/* remove_buddies */
	NULL,					/* add_permit */
	NULL,					/* add_deny */
	NULL,					/* rem_permit */
	NULL,					/* rem_deny */
	NULL,					/* set_permit_deny */
	NULL,					/* warn */
	NULL,					/* join_chat */
	NULL,					/* reject chat invite */
	NULL,					/* chat_invite */
	NULL,					/* chat_leave */
	NULL,					/* chat_whisper */
	NULL,					/* chat_send */
	NULL,					/* keepalive */
	NULL,					/* register_user */
	NULL,					/* get_cb_info */
	NULL,					/* get_cb_away */
	NULL,					/* alias_buddy */
	NULL,					/* group_buddy */
	NULL,					/* rename_group */
	NULL,					/* buddy_free */
	gaimwrite_close_convo,	   /* convo_closed */
	NULL,					/* normalize */
	NULL,					/* set_buddy_icon */
	NULL,					/* remove_group */
	NULL,					/* get_cb_real_name */
	NULL,					/* set_chat_topic */
	NULL,					/* find_blist_chat */
	NULL,					/* roomlist_get_list */
	NULL,					/* roomlist_cancel */
	NULL,					/* roomlist_expand_category */
	NULL,					/* can_receive_file */
	NULL					/* send_file */
};


static GaimPluginInfo info =
{
	GAIM_PLUGIN_MAGIC,
	GAIM_MAJOR_VERSION,
	GAIM_MINOR_VERSION,
	GAIM_PLUGIN_PROTOCOL,                             /**< type           */
	NULL,                                             /**< ui_requirement */
	0,                                                /**< flags          */
	NULL,                                             /**< dependencies   */
	GAIM_PRIORITY_DEFAULT,                            /**< priority       */

	"prpl-cbr-gaimwrite",                             /**< id             */
	"Gaim-Write",                                       /**< name           */
	"0.0.1",                                            /**< version        */
	                                                  /**  summary        */
	N_("Unix Write over SSH Protocol Plugin"),
	                                                  /**  description    */
	N_("Unix Write over SSH Protocol Plugin"),
	"Jeff Kaufman <cbr@sccs.swarthmore.edu>",              /**< author         */
	"http://sccs.swarthmore.edu/~cbr/gaim-write",         /**< homepage       */

	NULL,                                             /**< load           */
	NULL,                                             /**< unload         */
	NULL,                                             /**< destroy        */

	NULL,                                             /**< ui_info        */
	&prpl_info,                                       /**< extra_info     */
	NULL,
	NULL
};

static void
init_plugin(GaimPlugin *plugin)
{
	my_protocol = plugin;
}

GAIM_INIT_PLUGIN(null, init_plugin, info);
